/******************************************************************************/
/* "HINT" -- Hierarchical INTegration.                                        */
/* Copyright (C) 1994 by Iowa State University Research Foundation, Inc.      */
/*                                                                            */
/* This program is free software; you can redistribute it and/or modify       */
/* it under the terms of the GNU General Public License as published by       */
/* the Free Software Foundation.  You should have received a copy of the      */
/* GNU General Public License along with this program; if not, write to the   */
/* Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.    */
/*                                                                            */
/* Files needed for use:                                                      */
/*     * hint.c             ---- Driver source                                */
/*     * hkernel.c          ---- Kernel source                                */
/*     * hint.h             ---- General include file                         */
/*     * typedefs.h         ---- Include file for DSIZE and ISIZE             */
/*     * README             ---- These are the rules. Follow them!!!          */
/******************************************************************************/

#include       "hint.h"

#if defined(SGI)
#pragma no side effects ( Hint )
#endif

/******************************************************************************/
/*      Global Variables                                                      */
/******************************************************************************/
int         nproc;        /* Number of Processors                             */
#if defined(KSR)
int         tid;          /* Team id number for Kendall Square machines       */
#elif defined(BEOS)
thread_id   *threads;     /* array of thread pointers for BeOS                */
#endif

/******************************************************************************/
/* This sets the data file name.   */
/******************************************************************************/
#if defined(LDOUBLE)
	 char filnm[80]="LDOUBLE";
#elif defined(DOUBLE)
	 char filnm[80]="DOUBLE";
#elif defined(FLOAT)
	 char filnm[80]="FLOAT";
#elif defined(LONGLONG)
	 char filnm[80]="LONGLONG";
#elif defined(LONG)
	 char filnm[80]="LONG";
#elif defined(INT)
	 char filnm[80]="INT";
#elif defined(SHORT)
	 char filnm[80]="SHORT";
#elif defined(CHAR)
	 char filnm[80]="CHAR";
#endif


#if defined(BEOS) /* Need a couple of functions to pass the argumenst to Hint */
typedef struct
{
  int laps;
  DSIZE scx;
  DSIZE scy;
  ISIZE memry;
  int iproc;
  double *gs;
  RECT *rect;
  DSIZE *errs;
  ISIZE *ixes;
  ERROR *eflag;
} Hint_arg;

long spawn_Hint(void *arg);
long spawn_Hint(void *arg)
{
  Hint_arg a;
  thread_id id;
  long proc = (long)arg;
  int i;
  do
  {
    if (receive_data(&id, &a, sizeof(a)))
        return 0;
    for (i=a.laps;i>0;i--)
      Hint(a.scx, a.scy, a.memry, a.iproc, a.gs, 
        a.rect, a.errs, a.ixes, a.eflag);
    send_data(id, proc, &a, 0);
  }
  while(1);
}

__inline void Hint_args(Hint_arg *arg, int laps, DSIZE scx, DSIZE scy, ISIZE memry, int iproc, double *gs,
               RECT *rect, DSIZE *errs, ISIZE *ixes, ERROR *eflag);
__inline void Hint_args(Hint_arg *arg, int laps, DSIZE scx, DSIZE scy, ISIZE memry, int iproc, double *gs,
               RECT *rect, DSIZE *errs, ISIZE *ixes, ERROR *eflag)
{
  arg->laps = laps;
  arg->scx = scx;
  arg->scy = scy;
  arg->memry = memry;
  arg->iproc = iproc;
  arg->gs = gs; 
  arg->rect = rect;
  arg->errs = errs;
  arg->ixes = ixes;
  arg->eflag = eflag;
}
 
#endif


main (int argc, char *argv[])
{
    FILE    *curv;        /* Output file for QUIPS curve                      */
	char    numprocs[20]; /* String representation of the number of processors*/


    Speed   qdata[NSAMP]; /* Array to keep track of QUIPs and time            */

    ERROR   eflag;        /* Stopping condition returned from Hint            */

    ISIZE   imax,         /* Maximum representable index                      */
            itm, itm2,    /* Scratch values to find imax                      */
            n,            /* Number of goal subintervals                      */
            nscout;       /* Number of survey subintervals                    */

    volatile
    DSIZE   tm1;          /* Scratch to prevent use of registers              */

    DSIZE   dmax,         /* Maximum associative whole number                 */
            gamut,        /* Range of result from Hint, high minus low bound  */
            scx,          /* Scale factor for x values                        */
            scy,          /* Scale factor for y values                        */
            tm, tm2;      /* Scratch values to find dmax                      */

    double  delq,         /* Change in Quality                                */
			quips,        /* QUality Improvement Per Second                   */
            qpnet,        /* Net QUIPS; integral of QUIPS d(log t)            */
            qpeak,        /* Peak QUIPS attained                              */
            qprat,        /* Ratio of current QUIPS to peak QUIPS             */
            qptmp,        /* QUIPS temporary for calculating Net QUIPS        */
            t,            /* Time for a given trial                           */
            t0,           /* Starting time                                    */
            t1,           /* Ending time                                      */
            tdelta,       /* Timer resolution                                 */
            tscout;       /* Time for initial survey                          */

    QMSIZE  dbits,        /* Number of bits of accuracy for dmax              */
            ibits,        /* Number of bits for imax                          */
            i, j, k,      /* Loop counters                                    */
            laps,         /* Approximate number of laps in a trial            */
            memuse;       /* Amount of memory used, in bytes                  */

#if defined(BEOS)
    int iproc;
#endif

    printf("         _    _\n");
    printf("         |    |  _ _   _ _____ TM\n");
    printf("         |--  |  | |\\  | | | |\n");
    printf("         |  --|  | | \\ |   |\n");
    printf("         |    |  | |  \\|   |\n");
    printf("         ^    ^  ^ ^   ^   ^\n\n");
    printf("*** The  HINT  PERFORMANCE ANALYZER ***\n");
    printf("        Version 1.0  June 1994\n");
    printf("   John L. Gustafson & Quinn O. Snell\n");
    printf("     Scalable Computing Laboratory\n");
    printf("   236 Wilhelm, Iowa State University\n");
    printf("        Ames, Iowa    50011-3020\n");
    printf("             (515) 294 - 9294\n\n");
    printf("Copyright (C) 1994");
    printf(" Iowa State University Research Foundation, Inc.\n");
    printf("Please send results and questions to:   hint@scl.ameslab.gov\n");
    printf("When sending results please follow the form in README\n");
    printf("________________________________________________________\n");

/******************************************************************************/
/*  Determine Number of processors -                                          */
/*    Add in vendor specific code                                             */
/*    For example                     #elif defined(VENDOR)                   */
/*                                         nproc = vendor code                */
/******************************************************************************/
#if defined(SGI)
	if (argc == 1)
	  nproc = m_get_numprocs();
	else
	  nproc = atoi(argv[1]);
#elif defined(KSR)
    nproc = atoi(argv[1]);
    tid = pr_create_team(nproc);
#elif defined(BEOS)
    if (argc == 1)
    {
      system_info si; get_system_info(&si);
      nproc = si.cpu_count;
    }
    else
      nproc = atoi(argv[1]);
    if (threads = (thread_id *)malloc(nproc * sizeof(thread_id)))
    for (iproc = 0; iproc < nproc; iproc++)
    {
      threads[iproc] = spawn_thread( spawn_Hint, filnm, B_NORMAL_PRIORITY, (void *)iproc);
      resume_thread(threads[iproc]);
    }
#endif
    printf("Running on %d processors\n", nproc);


#ifdef DEBUG
    curv = stdout;
#else
    sprintf(numprocs,".%d",nproc);
	strcat(filnm, numprocs);
    if ((curv = fopen(filnm, "w")) == NULL)
    {
        printf("Could not open data file\n");
        exit(1);
    }
#endif

 /* Attempt to find timer resolution. Loop until the time changes.            */
    for (t0 = When(); ((t1 = When()) == t0););
    tdelta = t1 - t0;

 /* Find the largest associative whole number, dmax.                          */
    dbits = 0;
    tm = (DSIZE)1;
    tm2 = tm + tm;
    tm1 = tm2 + 1;
 /* Double until it fails to increment or it overflows.                       */
    while (((tm1 - tm2) == (DSIZE)1) && (tm2 > (DSIZE)0))
    {
        dbits++;
        tm  += tm;
        tm2 += tm2;
        tm1  = tm2 + 1;
    }
    dbits++;
 /* We use a grid of dmax + 1 squares, but this might overflow, so back off 1.*/
    dmax = tm + (tm - 1);
    printf("Apparent number of bits of accuracy: %d\n", dbits);
    printf("Maximum associative whole number:  %.0f\n",(double)dmax);

 /* Find the largest representable index number.                              */
    ibits = 0;
    itm = (ISIZE)1;
    itm2 = itm + itm;
 /* Double it until it overflows.                                             */
    while (itm2 > (ISIZE)0)
    {
        ibits++;
        itm  += itm;
        itm2 += itm2;
    }
    imax = itm;
    printf("Maximum number of bits of index: %d\n", ibits);
    printf("Maximum representable index:  %.0f\n", (double)imax);

 /* Calculate usable maxima, from whichever is most restrictive.              */
    if ((ibits + ibits) < dbits)
    {
        dmax = (DSIZE)imax * (DSIZE)imax - 1;
        dbits = ibits + ibits;
    }
    printf("\nIndex-limited data accuracy: %d bits\n", dbits);
    printf("Maximum usable whole number:  %.0f\n",(double)dmax);

 /* Haf the bits, biased downward, go to x.                                  */
    j = (dbits)/2;

 /* Now loop to take 2 to the appropriate power for grid dimensions.          */
    for (i = 0, scx = 1; i < j; scx += scx, i++);
    for (i = 0, scy = 1; i < dbits - j; scy += scy, i++);
    printf("Grid: %.0f wide by %.0f high.\n",(double)scx,(double)scy);

 /* Stop if there aren't enough subintervals to support scatter decomposition.*/
    if (nproc * NCHUNK  > scx)
    {
        printf("Too many processors. Only %.0f slots.\n",(double)scx);
        exit(0);
    }

 /* This loop is used as a survey.                                            */
    for (nscout = NCHUNK, laps = 3; nscout < scx / nproc / NCHUNK;)
    {
        t = Run(laps, &gamut, scx, scy, dmax, nscout, &eflag);

        if (eflag != NOERROR)
        {
            nscout /= 2;
            break;
        }
        else if ((t > RUNTM) && (eflag == NOERROR))
        {
            tscout = t;
            break;
        }
        else
        {
            tscout =  t;
            nscout *= 2;
            if (nscout > scx)
            {
                nscout /= 2;
                break;
            }
        }
    }

    if (tscout == 0)
    {
        fprintf(stderr, "Data type for %s is too small\n", argv[0]);
        exit(0);
    }
    if ((tscout < RUNTM) && (eflag == NOMEM))
        printf("Memory is not sufficient for > %3.1f s runs.\n", RUNTM);
    else if (tscout < RUNTM)
        printf("Precision is not sufficient for > %3.1f s runs\n", RUNTM);

 /* This loop is the main loop driver of the HINT kernel.                     */
    for (t = 0, i = 0, n = NCHUNK, qpeak = 0, qprat = 1;
    ((i < NSAMP) && (t < STOPTM) && (n < scx/nproc/NCHUNK) && (qprat > STOPRT));
        i++, n = ((QMSIZE)(n * ADVANCE) > n)? (n * ADVANCE) : n + 1)
    {
        printf(".");
        fflush(stdout);

     /* Use enough laps to fill RUNTM time, roughly.                          */
        laps = MAX(RUNTM * nscout / (tscout * n), 1);
        t = Run(laps, &gamut, scx, scy, dmax, n, &eflag);
        if (t == 0)
            t = tdelta;
		
        if (eflag != NOERROR)
            break;

     /* Calculate QUIPS. We must add one to dmax, but do it in steps.         */
     /* This to avoid overflow of dmax.                                       */
		delq = (double)dmax / gamut - 1;
        quips = delq / t + 1.0 / gamut / t;
        qdata[i].t = t;
        qdata[i].qp = quips;
        qdata[i].n = n;
        qpeak = MAX(qpeak, quips);
        qprat = quips / qpeak;
    }

    memuse = (QMSIZE)(qdata[i-1].n *
             (sizeof(RECT) + sizeof(DSIZE) + sizeof(ISIZE)) * nproc);
    if ((qprat > STOPRT) && (eflag == NOMEM))
        printf("\nThis run was memory limited at %d subintervals %d bytes.\n",
						 n, memuse);
    printf("\nDone with first pass. Now computing net QUIPS.\n");

    fprintf(curv,"%12.10f %10.0f %12.0f %12.0f %10d\n",
                qdata[i-1].t, qdata[i-1].qp, qdata[i-1].t * qdata[i-1].qp,
                (double)qdata[i-1].n, memuse);


 /* Now go backwards through the data to calculate net QUIPS, and filter data.*/
    for (qpnet = qdata[i-1].qp, j = i - 2; j >= 0; j--)
    {
     /* If more work takes less time, we need to rerun the case of less work. */
        for (k = 0; ((qdata[j+1].t < qdata[j].t) && (k < PATIENCE)); k++)
        {
            printf("#");
            fflush(stdout);
            laps  = MAX(RUNTM * nscout / (tscout * qdata[j].n), 1);
            t = Run(laps, &gamut, scx, scy, dmax, qdata[j].n, &eflag);
            if (t == 0)
                t = tdelta;
			
			if (t < 0)
			{
				printf("Time is less than zero %f\n", t);
				t = 1;
			}

			delq = (double)dmax / gamut - 1;
            quips = delq / t + 1.0 / gamut / t;
            qdata[j].t  = t;
            qdata[j].qp = quips;
        }
        if (qdata[j+1].t < qdata[j].t)
        {
            printf(" Couldn't eliminate interruptions. Forcing a time\n");
            qdata[j].t  = qdata[j+1].t;
			delq = (double)dmax / gamut - 1;
            qdata[j].qp = delq / qdata[j].t + 1.0 / gamut / qdata[j].t;
        }

        memuse = (QMSIZE)(qdata[j].n *
                      (sizeof(RECT) + sizeof(DSIZE) + sizeof(ISIZE)) * nproc);
        fprintf(curv,"%12.10f %10.0f %12.0f %12.0f %10d\n",
                        qdata[j].t, qdata[j].qp, qdata[j].t * qdata[j].qp,
                        (double)qdata[j].n, memuse);

     /* Now calculate the addition to the net QUIPS.                          */
     /* This is calculated as the sum of QUIPS(j) * (1 - time(j) / time(j+1)).*/
        qptmp = qdata[j].qp * qdata[j].t / qdata[j+1].t;
        qpnet += (qdata[j].qp - qptmp);
    }

/******************************************************************************/
/*  Vendor specific code to clean up.                                         */
/******************************************************************************/
#if defined(KSR)
    pr_destroy_team(tid);
#elif defined(BEOS)
    for (iproc = 0; iproc < nproc; iproc++)
    {
        long value;
        send_data(threads[iproc], 1, NULL, 0); /* this'll quit'em */
        wait_for_thread(threads[iproc], &value);
    }
    free(threads);
#endif

    printf("\nFinished with %f net QUIPS\n", qpnet);
    fclose(curv);
    return 0;
}


double
Run(int laps, 
    DSIZE *gamut, DSIZE scx, DSIZE scy, DSIZE dmax, ISIZE memry, ERROR *eflag)
{
    RECT   *(*rect)=NULL; /* Array for saving hierarchical information        */
    ISIZE  *(*ixes)=NULL; /* Array for indices of queued entries              */
    DSIZE  *(*errs)=NULL; /* Array of error values                            */

    int     i, j,         /* Loop counters                                    */
            iproc;        /* Loop counter for processors                      */

	volatile
    double  t0, t1, tm,   /* Time variables                                   */
            mint = 1e32;  /* Minimum time recorded for a given run            */

    double  gs[MXPROC*2];


/******************************************************************************/
/*  Memory Allocation                                                         */
/*    This should be done to allow each processor to use the fastest memory   */
/*    possible. The arrays rect[i], errs[i] and ixes[i] are private to each   */
/*    processor. You may want to do this on each processor in parallel.       */
/******************************************************************************/

 /* Allocate the memory for the array of pointers, one for each processor.    */
    rect = (RECT **)malloc((MSIZE)(nproc * sizeof(RECT *)));
    ixes = (ISIZE **)malloc((MSIZE)(nproc * sizeof(ISIZE *)));
    errs = (DSIZE **)malloc((MSIZE)(nproc * sizeof(DSIZE *)));

 /* If the memory is unavailable, free what is allocated and return.          */
    if ((rect == NULL) || (errs == NULL) || (ixes == NULL))
    {
        if (rect != NULL)
            free(rect);
        if (errs != NULL)
            free(errs);
        if (ixes != NULL)
            free(ixes);

        *eflag = NOMEM;
        return (-NOMEM);
    }

 /* Allocate the memory for each processor.                                   */
    for (i = 0; i < nproc; i++)
    {
        rect[i] = (RECT *)malloc((MSIZE)(memry * sizeof(RECT)));
        ixes[i] = (ISIZE *)malloc((MSIZE)(memry * sizeof(ISIZE) * 2));
        errs[i] = (DSIZE *)malloc((MSIZE)(memry * sizeof(DSIZE) * 2));
        if ((rect[i] == NULL) || (errs[i] == NULL) || (ixes[i] == NULL))
            *eflag = NOMEM;
    }

 /* If any of the arrays can't be allocated, we need to free all of them.     */
    if (*eflag == NOMEM)
    {
        for(i = 0; i < nproc; i++)
        {
            if (rect[i] != NULL)
                free(rect[i]);
            if (errs[i] != NULL)
                free(errs[i]);
            if (ixes[i] != NULL)
                free(ixes[i]);
        }
        free(rect);
        free(errs);
        free(ixes);
        return (-NOMEM);
    }
    
    *eflag = NOERROR;
    for (i = 0; i < NTRIAL; i++)
    {
        t0 = When();

#if !defined (BEOS)
        for (j = 0; j < laps; j++)
#endif
        {
            gs[HI] = gs[LO] = 0;

/******************************************************************************/
/*   Vendor Specific code to run the subroutine Hint in parallel.             */
/*   Add in the necessary code within a #elif defined(VENDOR) structure.      */
/******************************************************************************/
#if defined(SGI)
#pragma parallel
#pragma shared ( rect, errs, ixes, gs, eflag )
#pragma byvalue ( scx, scy, memry, nproc )
#pragma local ( iproc )
#pragma pfor iterate (iproc=0;nproc;1) 
            for (iproc = 0; iproc < nproc; iproc++)
            {
                Hint(scx, scy,  memry, iproc, &gs[2*iproc], 
                     rect[iproc], errs[iproc], ixes[iproc], eflag);
            }
			*gamut = 0;
			for (iproc = 0; iproc < nproc; iproc++)
				*gamut += gs[2*iproc] - gs[2*iproc+1];
				
#elif defined(KSR)
            pr_pcall(tid, Hint, copyargs(scx, scy, memry, iproc, gs,
                     rect[iproc], errs[iproc], ixes[iproc], eflag));
            *gamut = gs[HI] - gs[LO];
#elif defined(BEOS)
            for (iproc = 0; iproc < nproc; iproc++)
			{
				Hint_arg a;
				Hint_args(&a, laps, scx, scy, memry, iproc, &gs[2*iproc], 
				          rect[iproc], errs[iproc], ixes[iproc], eflag);
                send_data(threads[iproc], 0, &a, sizeof(a));
            }
			*gamut = 0;
            for (iproc = 0; iproc < nproc; iproc++)
            {
                thread_id id;
                int proc = receive_data(&id, NULL, 0);
                *gamut += gs[2*proc] - gs[2*proc+1];
            }
#endif
			if (*gamut < 0)
				printf("\nOOOOOPS gamut is %f [lo,hi]->[%f, %f]\n",
							*gamut, gs[LO], gs[HI]);
        }

        t1 = When();

     /* Track the minimum time thus far for this trial.                       */
        tm = (t1 - t0) / laps;
        mint = MIN(mint, tm);
    }

 /* Free up the memory.                                                       */
    for(i = 0; i < nproc; i++)
    {
        free(rect[i]);
        free(errs[i]);
        free(ixes[i]);
    }
    free(rect);
    free(errs);
    free(ixes);
    *eflag = NOERROR;
    return (mint);
}

/* Return the current time in seconds, using a double precision number.       */
double
When()
{
#if defined(BEOS)
    return (double)system_time()*1e-6;
#else
    struct timeval tp;
    gettimeofday(&tp, NULL);
    return ((double) tp.tv_sec + (double) tp.tv_usec * 1e-6);
#endif
}
