/******************************************************************************/
/* "HINT" -- Hierarchical INTegration.                                        */
/* Copyright (C) 1994 by Iowa State University Research Foundation, Inc.      */
/*                                                                            */
/* This program is free software; you can redistribute it and/or modify       */
/* it under the terms of the GNU General Public License as published by       */
/* the Free Software Foundation.  You should have received a copy of the      */
/* GNU General Public License along with this program; if not, write to the   */
/* Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.    */
/*                                                                            */
/* Files needed for use:                                                      */
/*     * hint.c             ---- Driver source                                */
/*     * hkernel.c          ---- Kernel source                                */
/*     * hint.h             ---- General include file                         */
/*     * typedefs.h         ---- Include file for DSIZE and ISIZE             */
/*     * README             ---- These are the rules. Follow them!!!          */
/******************************************************************************/

#include "hint.h"

/******************************************************************************/
/*      Global Variables                                                      */
/******************************************************************************/
extern int    nproc;     /* Number of Processors                              */

void
Hint(DSIZE scx, DSIZE scy, ISIZE mcnt, int iproc, double *gs,
               RECT *rect, DSIZE *errs, ISIZE *ixes, ERROR *eflag)
{
    DSIZE   errio,       /* Error value for io                                */
            errjo,       /* Error value for jo                                */
            sh,          /* Sum of areas, high bound                          */
            sl,          /* Sum of areas, low bound                           */
            tm,          /* Temporary value for computing function            */
            tm2,         /* Temporary value for doubling tm                   */
            tmerr;       /* Temporary error value for swaps                   */

    ISIZE   i,           /* Loop index                                        */
            inc, jnc,    /* Indices of queue positions                        */
            io,          /* Index of left child                               */
            iq,          /* Index of end of the sorted error queue            */
            it,          /* Iteration counter                                 */
            itmp,        /* Temporary                                         */
            jo,          /* Index of right child                              */
            ma,          /* Index of parent                                   */
            nchnk;       /* Number of initial partitions to work on           */
          

#if defined(KSR)
    iproc = pr_mid();
#endif

    sh = sl = 0;

 /* Initialize the first NCHUNK intervals for the scatter decomposition.      */
    for (iq = nchnk = 0; nchnk < NCHUNK; nchnk++, iq++)
    {
        rect[nchnk].dx = scx / NCHUNK / nproc;   
        rect[nchnk].xl = scx / NCHUNK * (iproc) / nproc +
                         (nchnk * nproc * rect[nchnk].dx);
        rect[nchnk].xr = rect[nchnk].xl + rect[nchnk].dx;

     /* Evaluate the function.                                                */
        tm  = (scx + rect[nchnk].xl);
        tm2 = (scy * (scx - ((rect[nchnk].xl == 0)? 1: rect[nchnk].xl)));
        itmp = tm2 / tm;
        rect[nchnk].fll = itmp;
        rect[nchnk].flh = rect[nchnk].fll + ((tm * rect[nchnk].fll) != tm2);

     /* Evaluate the function.                                                */
        tm  = (scx + rect[nchnk].xr);
        tm2 = (scy * (scx - rect[nchnk].xr));
        itmp = tm2 / tm;
        rect[nchnk].frl = itmp;

     /* Here we need boolean true == 1. Otherwise use if's.                   */
        rect[nchnk].frh = rect[nchnk].frl + ((tm * rect[nchnk].frl) != tm2);

        rect[nchnk].ahi = rect[nchnk].flh * rect[nchnk].dx;
        rect[nchnk].alo = rect[nchnk].frl * rect[nchnk].dx;
        tm = rect[nchnk].ahi - rect[nchnk].alo;

     /* Need to find the correct spot for this error to go into.              */
        for (i = nchnk; ((tm > errs[i-1]) && (i > 0)); i--)
        {
            errs[i] = errs[i-1];
            ixes[i] = ixes[i-1];
        }

        errs[i] = tm;
        ixes[i] = nchnk;      
        sh += rect[nchnk].ahi;
        sl += rect[nchnk].alo;
    }
 /* Need to back off iq so it points to the last entry in the queue.          */
    iq--;


    for (it = 0; ((it < mcnt - NCHUNK) && (it <= iq)); it++) 
    {
        io = ma = ixes[it];            /* Head of list has maximum error.     */
        jo = it + NCHUNK;              /* Find right child index.             */

        tm = rect[ma].dx;            
     /* Since dx is always a power of 2, it halves evenly.                    */
        rect[io].dx = rect[jo].dx = tm / (DSIZE)2;

     /* Right child gets right boundary.                                      */
        rect[jo].xr = rect[ma].xr;  
     /* New point in the middle is shared by child subintervals               */
        rect[io].xr  = rect[io].xl + 
                       rect[io].dx; 
        rect[jo].xl  = rect[io].xr;
     /* Right child gets right f value upper and lower bounds                 */
        rect[jo].frl = rect[ma].frl; 
        rect[jo].frh = rect[ma].frh;
     /* Note that the left child simply inherits much of its info from ma.    */

     /* This is the function evaluation.                                      */
        tm =  (scx + rect[io].xr);
        tm2 = (scy * (scx - rect[io].xr));
        itmp = tm2 / tm;
        rect[io].frl = itmp;

     /* Here we need boolean true == 1. Otherwise use if's.                   */
        rect[io].frh = rect[io].frl + ((tm * rect[io].frl) != tm2);
        rect[jo].fll = rect[io].frl;
        rect[jo].flh = rect[io].frh;

     /* Compute the left daughter error.                                      */
        tm = (rect[io].fll - rect[io].frh) * (rect[io].dx - (DSIZE)2);
        if (tm < (DSIZE)0)
            tm = (DSIZE)0;
        errio = (rect[io].flh - rect[io].frh  + 
                 rect[io].fll - rect[io].frl) * 
                (rect[io].dx - (DSIZE)1) - tm;

     /* Repeat for the right daughter.                                        */
        tm = (rect[jo].fll - rect[jo].frh) * (rect[jo].dx - (DSIZE)2);
        if (tm < (DSIZE)0)
            tm = (DSIZE)0;
        errjo = (rect[jo].flh - rect[jo].frh  + 
                 rect[jo].fll - rect[jo].frl) *  
                (rect[jo].dx - (DSIZE)1) - tm;

     /* Compute indices for io and jo on the queue.                           */
     /* This is done with a boolean. If boolean true is not 1, use if's.      */
        inc = (errio < errjo) + 1;
        jnc = 3 - inc;

     /* Put on both io and jo. If one is zero, we will not use it.            */
        errs[iq + inc] = errio;
        ixes[iq + inc] = io;
        errs[iq + jnc] = errjo;
        ixes[iq + jnc] = jo;

     /* Decide how much to increment iq. Again we need boolean true == 1.     */
        iq = iq + (errs[iq + 2] != 0) + 1;

     /* Remove parent sum contributions. Replace with child contributions.    */
        tm = rect[ma].alo;
        rect[io].alo = rect[io].frl * rect[io].dx;
        rect[jo].alo = rect[jo].frl * rect[jo].dx;
        sl -= tm;                    
        sl += rect[io].alo + rect[jo].alo; 
        tm = rect[ma].ahi;
        rect[io].ahi = rect[io].flh * rect[io].dx;
        rect[jo].ahi = rect[jo].flh * rect[jo].dx;
        sh -= tm;
        sh += rect[io].ahi + rect[jo].ahi;
		if (sh < sl)
			printf("\n%d: sh-> %f   and sl->%f\n",iproc,sh,sl);
    }

/******************************************************************************/
/*  Global sum operation.                                                     */
/*  Place vendor specific code in a #elif defined(VENDOR) structure. The      */
/*  individual values of sh and sl need to be summed into gs[HI] and gs[LO].  */
/******************************************************************************/
    gs[HI] = sh;
    gs[LO] = sl;

	if (gs[HI] < gs[LO])
		printf("%d: Here it is [%f, %f]\n", iproc, gs[LO], gs[HI]);

    if (it > iq)
        *eflag = DISCRET;
}
